
"""
Evaluate one or more benchmarks and upload the results to a Speedcenter server.
"""

from __future__ import division

if __name__ == '__main__':
    from speedcenter import main
    raise SystemExit(main())

from os import uname
from sys import executable, argv
from datetime import datetime
from subprocess import PIPE, Popen
from urllib import urlopen, urlencode

import twisted
from twisted.python.filepath import FilePath
from twisted.python.usage import UsageError

from all import allBenchmarkNames
from benchlib import BenchmarkOptions

# Unfortunately, benchmark name is the primary key for speedcenter
SPEEDCENTER_NAMES = {
    'tcp_connect': 'TCP Connections',
    'tcp_throughput': 'TCP Throughput',
    'ssh_connect': 'SSH Connections',
    'ssh_throughput': 'SSH Throughput',
    'ssl_connect': 'SSL Connections',
    'ssl_throughput': 'SSL Throughput',
    'sslbio_connect': 'SSL (Memory BIO) Connections',
    'sslbio_throughput': 'SSL (Memory BIO) Throughput',
    'int16receiver': 'int16strings',
    'lineonlyreceiver': 'only-lines',
    'linereceiver': 'lines',
    }


class SpeedcenterOptions(BenchmarkOptions):
    optParameters = [
        ('url', None, None, 'Location of Speedcenter to which to upload results.'),
        ]

    def postOptions(self):
        if not self['url']:
            raise UsageError("The Speedcenter URL must be provided.")



def reportEnvironment():
    revision = twisted.version.short().split('r', 1)[1]

    packageDirectory = FilePath(twisted.__file__).parent()

    try:
        import pysvn
    except ImportError:
        entries = packageDirectory.child('.svn').child('entries').getContent()
        lines = entries.splitlines()
        revision = lines[3]
        date = lines[9][:19].replace('T', ' ')
    else:
        client = pysvn.Client()
        [entry] = client.log(
            packageDirectory.path,
            pysvn.Revision(pysvn.opt_revision_kind.number, int(revision)),
            limit=1)
        date = str(datetime.fromtimestamp(entry['date']))

    return {
        'project': 'Twisted',
        'executable': executable,
        'environment': uname()[1].split('.')[0],
        'commitid': revision,
        'revision_date': date,
        'result_date': str(datetime.now()),
        }



def main():
    options = SpeedcenterOptions()
    try:
        options.parseOptions(argv[1:])
    except UsageError, e:
        raise SystemExit(str(e))

    duration = options['duration']
    iterations = options['iterations']

    results = {}

    for benchmark in allBenchmarkNames:
        print 'Running', benchmark, '...',
        child = Popen([
                executable, '-Wignore', benchmark + '.py',
                '--iterations', str(iterations),
                '--duration', str(duration),
                '--warmup', str(options['warmup']),
                '--format', '%(count)s'],
                                 stdout=PIPE)
        results[benchmark] = map(int, child.stdout.readlines())
        print 'Done'
        child.wait()

    environment = reportEnvironment()

    statsFile = open('stats', 'a')
    statsFile.write(str(environment) + '\n')
    statsFile.write(str(results) + '\n')
    statsFile.close()

    for (name, results) in sorted(results.items()):
        if not results:
            print '** Collected no results for', name
            continue

        rates = [count / duration for count in results]
        totalCount = sum(results)
        totalDuration = duration * iterations

        name = SPEEDCENTER_NAMES.get(name, name)
        stats = environment.copy()
        stats['benchmark'] = name
        stats['result_value'] = totalCount / totalDuration
        stats['min'] = min(rates)
        stats['max'] = max(rates)

        # Please excuse me.
        fObj = urlopen(options['url'], urlencode(stats))
        print name, fObj.read()
        fObj.close()
